"""
Meta File
"""

class OBORPCMeta(type):
    """
    Meta class used to construct RPC
    """
    __obor_registry__ = {}
    def __new__(mcs, name, bases, namespace, /, **kwargs):
        cls = super().__new__(mcs, name, bases, namespace, **kwargs)

        cls.__oborprocedures__ = {
            methodname for methodname, value in namespace.items()
            if getattr(value, "__isoborprocedure__", False)
        }
        OBORPCMeta.__obor_registry__[cls] = cls.__oborprocedures__

        return cls


class RPCBase(metaclass=OBORPCMeta): # pylint: disable=too-few-public-methods
    """
    Obor Base Class
    """
    def __repr__(self) -> str:
        return "<RPCBase(metaclass=OBORPCMeta)>"

    def __str__(self) -> str:
        return self.__repr__()



"""
FastAPI Server Builder
"""
import json
import asyncio
from enum import Enum
from typing import Optional, List, Dict, Union, Type, Any, Sequence, Callable

import jsonref
from pydantic import BaseModel
from fastapi import Request, Response, APIRouter, params
from fastapi.responses import JSONResponse
from fastapi.routing import BaseRoute, APIRoute, ASGIApp, Lifespan, Default, generate_unique_id

from ._server import ServerBuilder
from ..base.meta import RPCBase



class FastAPIServerBuilder(ServerBuilder):
    """
    Dedicated RPC Server Builder for FastAPI
    """
    def generate_model_schema(self, model: Optional[BaseModel]) -> Dict[str, Any]:
        """
        Generate pydantic model schema
        """
        try:
            if not model:
                return {}
            schema = json.dumps(model.model_json_schema())
            openapi_schema = json.loads(json.dumps(jsonref.loads(schema), indent=2))
            if "$defs" in openapi_schema:
                openapi_schema.pop("$defs")
            return openapi_schema
        except: # pylint: disable=bare-except
            return {}

    def generate_openapi_extra_body(self, class_name: str, method_name: str) -> Dict[str, Any]:
        """
        Generate OpenAPI Extra Body for Schema
        """
        # request schema
        model = self.model_maps[class_name][method_name][2]
        request_schema = self.generate_model_schema(model)

        # response schema
        model = self.model_maps[class_name][method_name][3]
        response_schema = self.generate_model_schema(model)

        extra = {
            "summary": f"{class_name}.{method_name}",
            "requestBody": {
                "content": {
                    "application/json": {
                        "schema": request_schema
                    }
                },
                "required": True,
            },
            "responses": {
                "200": {
                    "description": "Successful Response",
                    "content": {
                        "application/json": {
                            "schema": response_schema
                        }
                    }
                }
            }
        }
        return extra

    def create_remote_responder(
        self,
        instance: RPCBase,
        router: APIRouter,
        class_name: str,
        method_name: str,
        method: Callable
    ): # pylint: disable=too-many-positional-arguments
        @router.post(
            f"{router.prefix}/{class_name}/{method_name}",
            tags=[class_name],
            openapi_extra=self.generate_openapi_extra_body(class_name, method_name)
        )
        def rpc_function(request: Request):
            request_body = asyncio.run(request.body())
            if request_body:
                decoded_request_body = json.loads(request_body.decode())
                if isinstance(decoded_request_body, str):
                    body = json.loads(decoded_request_body)
                else:
                    body = decoded_request_body
            else:
                body = {}
            return self.dispatch_rpc_request(class_name, method_name, instance, method, body)

    def create_remote_responder_async(
        self,
        instance: RPCBase,
        router: APIRouter,
        class_name: str,
        method_name: str,
        method: Callable
    ): # pylint: disable=too-many-positional-arguments,too-many-arguments,too-many-arguments
        @router.post(
            f"{router.prefix}/{class_name}/{method_name}",
            tags=[class_name],
            openapi_extra=self.generate_openapi_extra_body(class_name, method_name)
        )
        async def rpc_function(request: Request):
            request_body = await request.body()
            if request_body:
                decoded_request_body = json.loads(request_body.decode())
                if isinstance(decoded_request_body, str):
                    body = json.loads(decoded_request_body)
                else:
                    body = decoded_request_body
            else:
                body = {}
            return await self.dispatch_rpc_request_async(class_name, method_name, instance, method, body)

    def build_router_from_instance(
        self,
        instance: RPCBase,
        *,
        prefix: str = "",
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        default_response_class: Type[Response] = Default(JSONResponse),
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        routes: Optional[List[BaseRoute]] = None,
        redirect_slashes: bool = True,
        default: Optional[ASGIApp] = None,
        dependency_overrides_provider: Optional[Any] = None,
        route_class: Type[APIRoute] = APIRoute,
        on_startup: Optional[Sequence[Callable[[], Any]]] = None,
        on_shutdown: Optional[Sequence[Callable[[], Any]]] = None,
        lifespan: Optional[Lifespan[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        generate_unique_id_function: Callable[[APIRoute], str] = Default(generate_unique_id),
        secure_build: bool = True,
    ): # pylint: disable=too-many-positional-arguments,too-many-arguments,too-many-locals
        """
        build FastAPI API Router from oborpc instance
        """
        router = APIRouter(
            prefix=prefix,
            tags=tags,
            dependencies=dependencies,
            default_response_class=default_response_class,
            responses=responses,
            callbacks=callbacks,
            routes=routes,
            redirect_slashes=redirect_slashes,
            default=default,
            dependency_overrides_provider=dependency_overrides_provider,
            route_class=route_class,
            on_startup=on_startup,
            on_shutdown=on_shutdown,
            lifespan=lifespan,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            generate_unique_id_function=generate_unique_id_function
        )

        self.setup_server_rpc(instance, router, secure_build)

        return router



"""
Client RPC Builder
"""
import inspect
import logging
import time
from typing import Any, Callable, Dict, Optional, Union

import httpx
import pydantic_core
from pydantic import BaseModel, create_model

from ..security import BASIC_AUTH_TOKEN
from ..exception import OBORPCBuildException, RPCCallException

# httpx log level
httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.WARNING)


class ClientBuilder:
    """
    Client Builder
    """
    __registered_base = set()

    def __init__(
        self,
        host: str,
        port: Optional[Union[str, int]] = None,
        timeout: Optional[float] = None,
        retry: Optional[int] = None,
        additional_headers: Optional[Dict[str, str]] = None,
        dynamic_headers_builder: Optional[Callable] = None,
    ): # pylint: disable=too-many-arguments,too-many-positional-arguments
        self.master_instances = []
        self.host = host
        self.port = port
        self.timeout = timeout or 1
        self.retry = retry or 0

        protocol = "http://"
        if self.check_has_protocol(host):
            protocol = ""

        self.base_url = f"{protocol}{host}"
        if port:
            self.base_url += f":{port}"

        # dynamic header
        self.dynamic_headers_builder = dynamic_headers_builder or (lambda: {})

        # static header
        headers = {
            "Authorization": f"Basic {BASIC_AUTH_TOKEN}",
            "Content-Type": "application/json"
        }
        if additional_headers:
            headers.update(additional_headers)
        self.headers = headers

        # request client
        self.request_client = httpx.Client(
            base_url=self.base_url,
            headers=headers
        )
        self.async_request_client = httpx.AsyncClient(
            base_url=self.base_url,
            headers=headers
        )

        # model map
        self.model_maps = {}

    def check_has_protocol(self, host: str):
        """
        Check whether the given host already defined with protocol or not
        """
        if host.startswith("http://"):
            return True
        if host.startswith("https://"):
            return True
        return False

    def check_registered_base(self, base: str):
        """
        Check whether the base RPC class is already built
        """
        if base in ClientBuilder.__registered_base:
            msg = f"Failed to build client RPC {base} : base class can only built once"
            raise OBORPCBuildException(msg)
        ClientBuilder.__registered_base.add(base)

    def create_remote_caller(
        self,
        class_name: str,
        method_name: str,
        url_prefix: str,
        timeout: float = None,
        retry: int = None
    ): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        create remote caller
        """
        def remote_call(*args, **kwargs):
            """
            remote call wrapper
            """
            start_time = time.time()
            try:
                url = f"{url_prefix}/{class_name}/{method_name}"
                headers = self.headers.copy()
                headers.update(self.dynamic_headers_builder())
                response = self.request_client.post(
                    url=url,
                    json=pydantic_core.to_jsonable_python({
                        "args": args[1:],
                        "kwargs": kwargs
                    }),
                    timeout=timeout if timeout is not None else self.timeout,
                    headers=headers
                )

                if not response:
                    msg = f"rpc call failed method={method_name}"
                    raise RPCCallException(msg)

                data = response.json().get("data")
                return self.convert_model_response(class_name, method_name, data)

            except Exception as e:
                _retry = retry if retry is not None else self.retry
                if _retry:
                    return remote_call(*args, **kwargs, retry=_retry-1)

                if isinstance(e, RPCCallException):
                    raise e
                msg = f"rpc call failed method={method_name} : {e}"
                raise RPCCallException(msg) from e

            finally:
                elapsed = f"{(time.time() - start_time) * 1000}:.2f"
                logging.debug("[RPC-Clientt] remote call take %s ms", elapsed)

        return remote_call

    def create_async_remote_caller(
        self,
        class_name: str,
        method_name: str,
        url_prefix: str,
        timeout: float = None,
        retry: int = None
    ): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        create async remote caller
        """
        async def async_remote_call(*args, **kwargs):
            """
            async remote call wrapper
            """
            start_time = time.time()
            try:
                data = pydantic_core.to_jsonable_python({"args": args[1:], "kwargs": kwargs})
                url = f"{url_prefix}/{class_name}/{method_name}"
                headers = self.headers.copy()
                headers.update(self.dynamic_headers_builder())
                response = await self.async_request_client.post(
                    url=url,
                    json=data,
                    timeout=timeout if timeout is not None else self.timeout,
                    headers=headers
                )

                if not response:
                    msg = f"rpc call failed method={method_name}"
                    raise RPCCallException(msg)

                data = response.json().get("data")
                return self.convert_model_response(class_name, method_name, data)

            except Exception as e:
                _retry = retry if retry is not None else self.retry
                if _retry:
                    return await async_remote_call(*args, **kwargs, retry=_retry-1)

                if isinstance(e, RPCCallException):
                    raise e
                msg = f"rpc call failed method={method_name} : {e}"
                raise RPCCallException(msg) from e

            finally:
                elapsed = f"{(time.time() - start_time) * 1000}:.2f"
                logging.debug("[RPC-Clientt] remote call take %s ms", elapsed)

        return async_remote_call

    def build_client_rpc(self, instance: object, url_prefix: str = ""):
        """
        Setup client rpc
        """
        _class = instance.__class__
        iterator_class = _class

        self.check_registered_base(_class)

        for (name, method) in inspect.getmembers(iterator_class, predicate=inspect.isfunction):
            if name not in iterator_class.__oborprocedures__:
                continue
            class_name = _class.__name__
            self.extract_models(class_name, name, method)
            setattr(_class, name, self.create_remote_caller(class_name, name, url_prefix))

    def build_async_client_rpc(self, instance: object, url_prefix: str = ""):
        """
        Setup async client rpc
        """
        _class = instance.__class__
        iterator_class = _class

        self.check_registered_base(_class)

        for (name, method) in inspect.getmembers(iterator_class, predicate=inspect.isfunction):
            if name not in iterator_class.__oborprocedures__:
                continue
            class_name = _class.__name__
            self.extract_models(class_name, name, method)
            setattr(_class, name, self.create_async_remote_caller(class_name, name, url_prefix))

    def extract_models(
        self,
        class_name: str,
        method_name: str,
        method: Callable
    ):
        """
        Extract pydantic model
        """
        if not class_name in self.model_maps:
            self.model_maps[class_name] = {}

        signature_params = inspect.signature(method).parameters
        params = {
            k: (
                v.annotation if v.annotation != inspect._empty else Any,
                v.default if v.default != inspect._empty else ...
            ) for k, v in signature_params.items()
        }

        signature_return = inspect.signature(method).return_annotation
        self.model_maps[class_name][method_name] = [
            create_model(f"{class_name}_{method_name}", **params),
            signature_return
        ]

    def convert_model_response(
        self,
        class_name: str,
        method_name: str,
        response: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Convert Pydantic Model Response
        """
        model, return_annotation = self.model_maps[class_name][method_name]
        try:
            if BaseModel.__subclasscheck__(return_annotation.__class__):
                return model.model_validate(response)
        except:
            pass
        return response



"""
Server Builder Base
"""
import inspect
from typing import Any, Callable, Dict, List

from pydantic import BaseModel, create_model

class ServerBuilder:
    """
    Server Builder
    """
    def __init__(self) -> None:
        self.model_maps: Dict[str, Dict[str, Any]] = {}

    def create_remote_responder(self, instance, router, class_name, method_name, method): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        Remote RPC Request Responder
        """
        raise NotImplementedError("method should be overridden")

    def create_remote_responder_async(self, instance, router, class_name, method_name, method): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        Remote RPC Request Responder Async
        """
        raise NotImplementedError("method should be overridden")

    def dispatch_rpc_request(self, class_name, method_name, instance, method, body): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        Dispatch RPC Request
        """
        kwargs = self.construct_model_object(class_name, method_name, body)
        res = method(instance, **kwargs)
        return {"data": self.convert_model_response(res)}

    async def dispatch_rpc_request_async(self, class_name, method_name, instance, method, body): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        Dispatch RPC Request
        """
        kwargs = self.construct_model_object(class_name, method_name, body)
        res = await method(instance, **kwargs)
        return {"data": self.convert_model_response(res)}

    def setup_server_rpc(self, instance: object, router, secure_build: bool = True):
        """
        Setup RPC Server
        """
        _class = instance.__class__
        method_map = { # pylint: disable=unnecessary-comprehension
            name: method for (name, method) in inspect.getmembers(
                _class, predicate=inspect.isfunction
            )
        }

        iterator_class = instance.__class__.__base__
        iterator_method_map = { # pylint: disable=unnecessary-comprehension
            name: method for (name, method) in inspect.getmembers(
                iterator_class, predicate=inspect.isfunction
            )
        }

        for (name, method) in inspect.getmembers(iterator_class, predicate=inspect.isfunction):
            if name not in iterator_class.__oborprocedures__:
                continue

            # validate
            method = method_map.get(name)
            iterator_method = iterator_method_map.get(name)
            if secure_build:
                self.validate_implementation(
                    name,
                    method,
                    _class,
                    iterator_method,
                    iterator_class
                )

            # build router
            class_name = iterator_class.__name__
            self.extract_models(class_name, name, method)
            if inspect.iscoroutinefunction(method):
                self.create_remote_responder_async(instance, router, class_name, name, method)
            else:
                self.create_remote_responder(instance, router, class_name, name, method)

    def validate_implementation(
        self,
        method_name: str,
        implementation_method: Callable,
        implementation_class: object,
        origin_method: Callable,
        origin_class: object,
    ): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        Validate implementation of RPC super class
        """
        # validate implementation: check overridden procedure
        method_str = str(implementation_method)
        method_origin = method_str[9:method_str.find(" at 0x")].split(".")[0].strip()
        implementation_origin = str(implementation_class)[8:-2].split(".")[-1].strip()
        err = f"Unable to build. Procedure `{implementation_origin}.{method_name}()` is not implemented"
        assert method_origin == implementation_origin, err

        # validate implementation: check procedure has the same callable type
        is_implementation_coroutine = inspect.iscoroutinefunction(implementation_method)
        is_origin_coroutine = inspect.iscoroutinefunction(origin_method)
        callable_type = ["def", "async def"]
        iterator_origin = str(origin_class)[8:-2].split(".")[-1].strip()
        err = (
            f"Unable to build. Procedure `{implementation_origin}.{method_name}()` "
            f"is implemented as `{callable_type[int(is_implementation_coroutine)]}`. "
            f"While the origin `{iterator_origin}.{method_name}()` is defined as "
            f"`{callable_type[int(is_origin_coroutine)]}`."
        )
        assert is_implementation_coroutine == is_origin_coroutine, err

    def extract_models(self, class_name: str, method_name: str, method: Callable):
        """
        Extract pydantic models from method signature for both spec and returns
        """
        if not class_name in self.model_maps:
            self.model_maps[class_name] = {}

        signature = inspect.signature(method)

        # request signature
        request_params = {
            k: (
                v.annotation if v.annotation != inspect._empty else Any,
                v.default if v.default != inspect._empty else ...
            ) for i, (k, v) in enumerate(signature.parameters.items())
            if i != 0
        }
        kwargs_model = create_model(f"{class_name}_{method_name}_request_kwargs", **request_params)
        request_premodel = {
            "args": (List[Any], []),
            "kwargs": (kwargs_model, ...)
        }
        request_model = create_model(f"{class_name}_{method_name}_request", **request_premodel)

        # response signature
        return_annot = signature.return_annotation
        return_annot = return_annot if return_annot != inspect._empty else Any
        response_params = {"data": (return_annot, ...)}

        self.model_maps[class_name][method_name] = [
            list(signature.parameters.keys())[1:],
            kwargs_model,
            request_model,
            create_model(f"{class_name}_{method_name}_reponse", **response_params)
        ]

    def construct_model_object(
        self,
        class_name: str,
        method_name: str,
        body: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Construct pydantic object with the declared model and the given request body
        """
        arg_keys, kwargs_model, _, _ = self.model_maps[class_name][method_name]
        args = body.get("args", [])
        kwargs = body.get("kwargs", {})
        for i, arg in enumerate(args):
            kwargs[arg_keys[i]] = arg
        return vars(kwargs_model.model_validate(kwargs))

    def convert_model_response(self, response: Dict[str, Any]) -> Dict[str, Any]:
        """
        Convert response into pydantic model if possible
        """
        if BaseModel.__subclasscheck__(response.__class__):
            return response.model_dump()
        return response



"""
OBORPC Base Builder
"""
from ..exception import OBORPCBuildException

class OBORBuilder():
    """
    OBORPC Builder Class
    """
    __registered_base = set()

    def __init__(self, host, port=None, timeout=1, retry=0) -> None:
        self.master_instances = []
        self.host = host
        self.port = port
        self.timeout = timeout
        self.retry = retry

        protocol = "http://"
        if self.check_has_protocol(host):
            protocol = ""

        self.base_url = f"{protocol}{host}"
        if port:
            self.base_url += f":{port}"

    def check_has_protocol(self, host: str):
        """
        Check whether the given host already defined with protocol or not
        """
        if host.startswith("http://"):
            return True
        if host.startswith("https://"):
            return True
        return False

    def check_registered_base(self, base: str):
        """
        Check whether the base RPC class is already built
        """
        if base in OBORBuilder.__registered_base:
            msg = f"Failed to build client RPC {base} : base class can only built once"
            raise OBORPCBuildException(msg)
        OBORBuilder.__registered_base.add(base)



"""
OBORPC Exceptions
"""

class OBORPCBuildException(Exception):
    """
    Build Related Exceptions
    """
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

class RPCCallException(Exception):
    """
    Any Exception during RPC Call
    """
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



"""
Security Related Utility
"""
import httpx

BASIC_AUTH_TOKEN = "44a3f281214d481b8408e7b2355968d1"


def validate_rpc_response(response: httpx.Response):
    """
    Validate rpc response with defined security basic token
    """
    if response.status_code != 200:
        return False

    try:
        valid_content_type = (
            response.headers.get("content-type", "").lower()
            == "application/json"
        )
        valid_token = (
            response.headers.get("authorization", "").lower()
            == f"basic {BASIC_AUTH_TOKEN}"
        )
        return valid_content_type and valid_token
    except:
        return False



"""
Flask Server Builder
"""
import functools
import json
import os
from typing import Callable, Union
from flask import request as flask_request, Blueprint
from ._server import ServerBuilder
from ..base.meta import RPCBase

class FlaskServerBuilder(ServerBuilder):
    """
    Dedicated RPC Server Builder for Flask
    """
    def create_remote_responder(
        self,
        instance: RPCBase,
        router: Blueprint,
        class_name: str,
        method_name: str,
        method: Callable
    ): # pylint: disable=too-many-arguments,too-many-positional-arguments
        def create_modified_func():
            @functools.wraps(method)
            def modified_func():
                request_body = flask_request.get_json()
                body = json.loads(flask_request.get_json()) if request_body else {}
                return self.dispatch_rpc_request(class_name, method_name, instance, method, body)
            return modified_func
        router.post(f"{router.url_prefix or ''}/{class_name}/{method_name}")(create_modified_func())

    def build_blueprint_from_instance(
        self,
        instance: RPCBase,
        blueprint_name: str,
        import_name: str,
        static_folder: Union[str, os.PathLike, None] = None,
        static_url_path: Union[str, None] = None,
        template_folder: Union[str, os.PathLike, None] = None,
        url_prefix: Union[str, None] = None,
        subdomain: Union[str, None] = None,
        url_defaults: Union[dict, None] = None,
        root_path: Union[str, None] = None,
        cli_group: Union[str, None] = object(),
        secure_build: bool = True,
    ): # pylint: disable=too-many-arguments,too-many-positional-arguments
        """
        build Flask blueprint from oborpc instance
        """
        blueprint = Blueprint(
            blueprint_name,
            import_name,
            static_folder,
            static_url_path,
            template_folder,
            url_prefix,
            subdomain,
            url_defaults,
            root_path,
            cli_group
        )

        self.setup_server_rpc(instance, blueprint, secure_build)

        return blueprint



"""
Setup script
"""
from pathlib import Path
from setuptools import setup, find_packages

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

VERSION = "0.2.5"
DESCRIPTION = "An easy setup object oriented RPC. Built-in setup for FastAPI and Flask"

# Setting up
setup(
    name="oborpc",
    version=VERSION,
    author="danangjoyoo (Agus Danangjoyo)",
    author_email="<agus.danangjoyo.blog@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=["httpx", "pydantic", "jsonref"],
    keywords=["fastapi", "flask", "rpc", "OOP"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Typing :: Typed"
    ],
    url="https://github.com/Danangjoyoo/oborpc"
)



"""
OBORPC Decorator
"""
import inspect
from typing import Any, Callable, TypeVar

DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])

def procedure(fun: Callable) -> DecoratedCallable:
    """
    Marks the method with this decorator to make it available
    for RPC within the class with direct inheritance with `RPCBase`.
    To make it works also make sure the method is overridden
    in the server class inheritance of the base class

    from oborpc.base import meta
    from oborpc.decorator import procedure

    class Calculator(meta.RPCBase):
        @procedure
        def add(self, a, b):
            pass

    class CalculatorServer(Calculator):
        def tambah(self, a, b):
            return a + b
    """
    if not inspect.isfunction(fun):
        raise TypeError("can only applied for function or method")
    fun.__isoborprocedure__ = True
    return fun
